package ru.dbotthepony.mc.otm.compat.jei

import it.unimi.dsi.fastutil.ints.Int2IntArrayMap
import it.unimi.dsi.fastutil.ints.Int2IntFunction
import it.unimi.dsi.fastutil.ints.IntArrayList
import it.unimi.dsi.fastutil.ints.IntArraySet
import it.unimi.dsi.fastutil.objects.Object2IntArrayMap
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import mezz.jei.api.IModPlugin
import mezz.jei.api.JeiPlugin
import mezz.jei.api.constants.RecipeTypes
import mezz.jei.api.gui.handlers.IGuiContainerHandler
import mezz.jei.api.gui.ingredient.IRecipeSlotView
import mezz.jei.api.gui.ingredient.IRecipeSlotsView
import mezz.jei.api.helpers.IJeiHelpers
import mezz.jei.api.recipe.RecipeIngredientRole
import mezz.jei.api.recipe.transfer.IRecipeTransferError
import mezz.jei.api.recipe.transfer.IRecipeTransferHandler
import mezz.jei.api.registration.IGuiHandlerRegistration
import mezz.jei.api.registration.IRecipeCatalystRegistration
import mezz.jei.api.registration.IRecipeCategoryRegistration
import mezz.jei.api.registration.IRecipeRegistration
import mezz.jei.api.registration.IRecipeTransferRegistration
import net.minecraft.client.renderer.Rect2i
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.MenuType
import net.minecraft.world.inventory.Slot
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.crafting.CraftingRecipe
import net.minecraft.world.item.crafting.RecipeType
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.menu.ExoSuitInventoryMenu
import ru.dbotthepony.mc.otm.network.MenuNetworkChannel
import ru.dbotthepony.mc.otm.network.MoveMultipleItemsPacket
import ru.dbotthepony.mc.otm.recipe.EnergyContainerRecipe
import ru.dbotthepony.mc.otm.registry.MItems
import ru.dbotthepony.mc.otm.registry.MRecipes
import java.util.*
import java.util.function.BiFunction
import java.util.stream.Collectors
import kotlin.collections.ArrayList
import kotlin.properties.Delegates

var isJeiLoaded = false
	private set

@JeiPlugin
@Suppress("unused")
class JEIPlugin : IModPlugin {
	companion object {
		private val LOCATION = ResourceLocation(OverdriveThatMatters.MOD_ID, "jei_plugin")

		var helpers: IJeiHelpers by Delegates.notNull()
			private set
	}

	init {
		isJeiLoaded = true
	}

	override fun getPluginUid(): ResourceLocation {
		return LOCATION
	}

	override fun registerRecipeCatalysts(registration: IRecipeCatalystRegistration) {
		registration.addRecipeCatalyst(ItemStack(MItems.CHEMICAL_GENERATOR), RecipeTypes.FUELING)
		registration.addRecipeCatalyst(ItemStack(MItems.PLATE_PRESS), PlatePressRecipeCategory.recipeType)
	}

	override fun registerCategories(registration: IRecipeCategoryRegistration) {
		helpers = registration.jeiHelpers
		registration.addRecipeCategories(PlatePressRecipeCategory)
	}

	override fun registerRecipes(registration: IRecipeRegistration) {
		val level = minecraft.level ?: throw NullPointerException("No ClientLevel. OLOLOLOLOLOLO")

		registration.addRecipes(PlatePressRecipeCategory.recipeType, level.recipeManager.getAllRecipesFor(MRecipes.PLATE_PRESS).filter { !it.isIncomplete })
		registration.addRecipes(RecipeTypes.CRAFTING, level.recipeManager.getAllRecipesFor(RecipeType.CRAFTING).filterIsInstance<EnergyContainerRecipe>().map { it.parent })
	}

	override fun registerRecipeTransferHandlers(registration: IRecipeTransferRegistration) {
		val helper = registration.transferHelper

		registration.addRecipeTransferHandler(object : IRecipeTransferHandler<ExoSuitInventoryMenu, CraftingRecipe> {
			override fun getContainerClass(): Class<out ExoSuitInventoryMenu> {
				return ExoSuitInventoryMenu::class.java
			}

			override fun getMenuType(): Optional<MenuType<ExoSuitInventoryMenu>> {
				return Optional.empty()
			}

			override fun getRecipeType(): mezz.jei.api.recipe.RecipeType<CraftingRecipe> {
				return RecipeTypes.CRAFTING
			}

			private val validSlots = IntArraySet().also {
				it.add(0)
				it.add(1)
				it.add(3)
				it.add(4)
			}

			private val directMap = Int2IntArrayMap().also {
				for (i in 0 .. 8) {
					it.put(i, i)
				}
			}

			private val smallMap = Int2IntArrayMap().also {
				var i = 0
				it.put(0, i++)
				it.put(1, i++)
				it.put(3, i++)
				it.put(4, i)
			}

			// i wonder why Mezz didn't expose IConnectionToServer
			// because without it there is a lot of code duplication!
			override fun transferRecipe(
				container: ExoSuitInventoryMenu,
				recipe: CraftingRecipe,
				recipeSlots: IRecipeSlotsView,
				player: Player,
				maxTransfer: Boolean,
				doTransfer: Boolean
			): IRecipeTransferError? {
				val inputs = recipeSlots.getSlotViews(RecipeIngredientRole.INPUT)

				if (inputs.size != 9) {
					return IRecipeTransferError { IRecipeTransferError.Type.INTERNAL }
				}

				if (container.craftingSlots.size != 9) {
					for ((i, ingredient) in inputs.withIndex()) {
						if (!ingredient.isEmpty && i !in validSlots) {
							return helper.createUserErrorWithTooltip(TranslatableComponent("jei.tooltip.error.recipe.transfer.too.large.player.inventory"))
						}
					}
				}

				val missingItems = inputs.stream()
					.filter { !it.isEmpty }
					.filter { it.itemStacks.noneMatch { recipeItem -> container.allAccessibleSlots.any { it.mayPickup(player) && it.item.item === recipeItem.item } } }
					.collect(Collectors.toList())

				if (missingItems.isNotEmpty()) {
					return helper.createUserErrorForMissingSlots(TranslatableComponent("jei.tooltip.error.recipe.transfer.missing"), missingItems)
				}

				val emptySlots = container.allAccessibleSlots.stream().filter { it.item.isEmpty }.toList()
				val filledCraftingSlots = container.craftingSlots.stream().filter { !it.item.isEmpty }.count()

				if (emptySlots.size < filledCraftingSlots) {
					return helper.createUserErrorWithTooltip(TranslatableComponent("jei.tooltip.error.recipe.transfer.inventory.full"))
				}

				if (!doTransfer) {
					return null
				}

				val desiredItems = inputs.stream()
					.filter { !it.isEmpty }
					.flatMap { it.itemStacks }
					.map { it.item }
					.distinct()
					.collect(Collectors.toSet())

				val sources = Object2ObjectArrayMap<Item, MutableList<Slot>>()

				for (slot in container.allAccessibleSlots) {
					if (!slot.item.isEmpty && slot.mayPickup(player) && slot.item.item in desiredItems) {
						sources.computeIfAbsent(slot.item.item, Object2ObjectFunction { ArrayList() }).add(slot)
					}
				}

				val bestSources = Array(inputs.size) {
					val ingredient = inputs[it]

					if (ingredient.isEmpty) {
						kotlin.collections.ArrayDeque()
					} else {
						var bestMatch: List<Slot>? = null
						var bestMatchSize = 0

						for (item in ingredient.itemStacks) {
							val slots = sources[item.item]

							if (slots != null) {
								var thisSize = 0
								for (slot in slots) thisSize += slot.item.count

								if (bestMatchSize < thisSize) {
									bestMatchSize = thisSize
									bestMatch = slots
								}
							}
						}

						val bestMatchNN = bestMatch ?: throw NullPointerException("Could not find best matching set of slots for recipe autofill (this exception should be never reachable)")
						val trulyBest: List<Slot>

						if (bestMatchNN.size <= 99) {
							trulyBest = bestMatchNN
						} else {
							trulyBest = ArrayList<Slot>(99).also { for (i in 0 .. 39) it.add(bestMatchNN[i]) }
						}

						kotlin.collections.ArrayDeque<Slot>(trulyBest.size).also {
							it.addAll(trulyBest)

							it.sortWith { a, b ->
								a.item.count.compareTo(b.item.count)
							}
						}
					}
				}

				val actions = ArrayList<MoveMultipleItemsPacket.Move>()

				val nextEmpty = emptySlots.iterator()
				val emptiedSlots = IntArraySet()

				// clear crafting slots
				for (slot in container.craftingSlots) {
					if (!slot.item.isEmpty) {
						val emptySlot = nextEmpty.next()
						actions.add(MoveMultipleItemsPacket.Move(setOf(slot.index), emptySlot.index, slot.item.count))
						emptiedSlots.add(slot.index)

						for (sourceList in bestSources) {
							if (sourceList.isNotEmpty() && sourceList.first().item.item === slot.item.item) {
								sourceList.addFirst(emptySlot)
							}
						}
					}
				}

				val pieWishers = Object2IntArrayMap<Item>()

				for (slots in bestSources) {
					if (slots.isNotEmpty()) {
						pieWishers.compute(slots.first().item.item) { _, count -> (count ?: 0) + 1 }
					}
				}

				var transferCount = if (maxTransfer) 64 else 1

				if (transferCount != 1) {
					for ((i, ingredient) in inputs.withIndex()) {
						if (!ingredient.isEmpty) {
							var maxTransferThis = 0
							for (slot in bestSources[i]) maxTransferThis += slot.item.count
							transferCount = transferCount.coerceAtMost(maxTransferThis / pieWishers.getInt(bestSources[i].first().item.item))
						}
					}
				}

				val remap = if (container.craftingSlots.size == 9) directMap else smallMap

				for ((i, ingredient) in inputs.withIndex()) {
					if (!ingredient.isEmpty) {
						actions.add(MoveMultipleItemsPacket.Move(
							IntArrayList().also { for (slot in bestSources[i]) if (slot.index !in emptiedSlots) it.add(slot.index) },
							container.craftingSlots[remap[i]].index,
							transferCount))
					}
				}

				val packetA = MoveMultipleItemsPacket(actions, mapOf())
				val result = packetA.execute(player, container)
				val packetB = MoveMultipleItemsPacket(actions, result)

				MenuNetworkChannel.sendToServer(packetB)

				return null
			}
		}, RecipeTypes.CRAFTING)
	}

	override fun registerGuiHandlers(registration: IGuiHandlerRegistration) {
		registration.addGenericGuiContainerHandler(MatteryScreen::class.java, object : IGuiContainerHandler<MatteryScreen<*>> {
			override fun getGuiExtraAreas(containerScreen: MatteryScreen<*>): MutableList<Rect2i> {
				return containerScreen.panelsView.stream().map { it.calculateAbsoluteRectangles() }.flatMap { it.stream() }.map { it.toIntRect() }.collect(Collectors.toList())
			}

			override fun getIngredientUnderMouse(
				containerScreen: MatteryScreen<*>,
				mouseX: Double,
				mouseY: Double
			): Any? {
				return containerScreen.panelsView.stream().map { it.findItemStack(mouseX.toFloat(), mouseY.toFloat(), ignoreMouseInputLock = true) }.filter { it.first }.findAny().orElse(null)?.second
			}
		})
	}
}
